//****************************************************************************
//**
//**    VID_WIN.CPP
//**    Windows DirectDraw
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32
#include "sys_defs.h"
#include "vid_win.h"

#include <ddraw.h>
//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define VIDW_MAXDEVICES  32		// maximum number of devices to use
#define VIDW_MAXMODES    1024	// maximum number of modes to use

#define VIDW_MAXIMAGEHANDLES     1024    // max number of simultaneous images

enum
{
    VIF_MASKING     = 0x00000001
};

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================

// video device structure, represents a directdraw device
typedef struct
{
    dwbool isNullGuid;
    GUID deviceGuid;
    char name[248];
    char description[256];
    DDCAPS devCaps;
    DDCAPS helCaps;
} vidw_device_t;

// video mode structure
typedef struct
{
    vidw_device_t *device;
    dword width;
    dword height;
    dword bitDepth;
    dword refresh;
    DDSURFACEDESC surfDesc;
} vidw_mode_t;

// image structure
typedef struct
{
    LPDIRECTDRAWSURFACE3 ddMemSurface, ddVidSurface;
    dword width, height, bitDepth;
    dword flags; // VIF_ flags
} vidw_image_t;

//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
// directdraw variables
static LPDIRECTDRAW2 vidw_lpDD2 = NULL;
static LPDIRECTDRAWSURFACE3 vidw_lpDDSPrimary = NULL;
static LPDIRECTDRAWSURFACE3 vidw_lpDDSBack = NULL;
static LPDIRECTDRAWPALETTE vidw_lpDDPalette = NULL;

// device and mode pools
static dword vidw_numDevices = 0;
static vidw_device_t vidw_Devices[VIDW_MAXDEVICES];
static dword vidw_numModes = 0;
static vidw_mode_t vidw_Modes[VIDW_MAXMODES];

// status values
static vidw_device_t* vidw_currentDevice = NULL;
static vidw_mode_t* vidw_currentMode = NULL;
static dword vidw_ScreenWidth, vidw_ScreenHeight, vidw_ScreenBitDepth;
//static dword vidw_ScreenPitch;
//static byte* vidw_ScreenPtr;
static byte vidw_Palette[768];

// image pool
static vidw_image_t vidw_imagePool[VIDW_MAXIMAGEHANDLES];

//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static void ModeShutdown();
static void DeviceShutdown();
static void ActivateDevice(vidw_device_t* device);
static void ActivateMode(vidw_mode_t* mode);
static vidw_image_t* ImageForId(dword id);
static dword IdForImage(vidw_image_t* img);
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================

// shut down the current video mode
static void ModeShutdown()
{
	// destroy surfaces and palette created at mode activation
/*
	if (vidw_lpDDSOffscreen)
	{
		vidw_lpDDSOffscreen->Unlock(vidw_ScreenPtr);
		vidw_lpDDSOffscreen->Release();
		vidw_lpDDSOffscreen = NULL;
	}
*/
	if (vidw_lpDDSBack)
	{
		vidw_lpDDSBack->Release();
		vidw_lpDDSBack = NULL;
	}
	if (vidw_lpDDSPrimary)
	{
		vidw_lpDDSPrimary->Release();
		vidw_lpDDSPrimary = NULL;
	}	
	if (vidw_lpDDPalette)
	{
		vidw_lpDDPalette->Release();
		vidw_lpDDPalette = NULL;
	}

	// set the cooperative level back
	if (vidw_lpDD2)
		vidw_lpDD2->SetCooperativeLevel(sysw_hwnd, DDSCL_NORMAL);

    vidw_currentMode = NULL;
}

// shut down the current device
static void DeviceShutdown()
{
	// release the directdraw interface
	if (vidw_lpDD2)
	{
		vidw_lpDD2->Release();
		vidw_lpDD2 = NULL;
	}

    vidw_currentDevice = NULL;
}

// activate a directdraw device
static void ActivateDevice(vidw_device_t* device)
{
    LPDIRECTDRAW lpDD;
    HRESULT hr;

    // if the device isn't changing, skip out
	if (vidw_currentDevice == device)
		return;

    // shut the current device down
	DeviceShutdown();
    
    // assign the new active device
    vidw_currentDevice = device;

    // don't start a new device when null is passed in
    if (!device)
        return;
    
    // create the new device interface
    if (device->isNullGuid)
        hr = DirectDrawCreate(NULL, &lpDD, NULL);
    else
        hr = DirectDrawCreate(&device->deviceGuid, &lpDD, NULL);
    if (hr != DD_OK)
        SYS_Errorf("ActivateDevice: DirectDrawCreate on device %s(%s)",
            device->name, device->description);
    
	if (lpDD->QueryInterface(IID_IDirectDraw2, (LPVOID*)&vidw_lpDD2) != DD_OK)
        SYS_Errorf("ActivateDevice: QueryInterface lpDD2 on device %s(%s)",
            device->name, device->description);
    
    lpDD->Release();
}

// activate a video mode
static void ActivateMode(vidw_mode_t* mode)
{
    DDSURFACEDESC ddsdesc;
    DDSCAPS ddscaps;
    PALETTEENTRY pal[256];
    dword i;
    byte* bptr;
    PALETTEENTRY* palptr;
    HRESULT hr;
    LPDIRECTDRAWSURFACE initialPrimary, initialBack;

    // if the mode isn't changing, skip out
	if (vidw_currentMode == mode)
		return;

	// shut the current mode down
	ModeShutdown();

    // switch devices as necessary
    if (!mode)
        ActivateDevice(NULL);
    else
        ActivateDevice(mode->device);

    // assign the new active mode
    vidw_currentMode = mode;

    // don't set a new mode when null is passed in
    if (!mode)
        return;

    // change the cooperative level
    if ((hr = vidw_lpDD2->SetCooperativeLevel(sysw_hwnd,
        DDSCL_FULLSCREEN|DDSCL_EXCLUSIVE|DDSCL_ALLOWREBOOT)) != DD_OK)
    {
        SYS_Errorf("ActivateMode: SetCooperativeLevel on device %s(%s)",
            mode->device->name, mode->device->description);
    }

    // change the display mode
    if (vidw_lpDD2->SetDisplayMode(mode->width, mode->height,
        mode->bitDepth, 0, 0) != DD_OK)
    {
        SYS_Errorf("ActivateMode: SetDisplayMode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);
    }

    // get the primary surface
	memset(&ddsdesc, 0, sizeof(ddsdesc));
	ddsdesc.dwSize = sizeof(ddsdesc);
	ddsdesc.dwFlags = 0;
	ddsdesc.dwBackBufferCount = 1; ddsdesc.dwFlags |= DDSD_BACKBUFFERCOUNT;
	ddsdesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP; ddsdesc.dwFlags |= DDSD_CAPS;

    if (vidw_lpDD2->CreateSurface(&ddsdesc, &initialPrimary, NULL) != DD_OK)
        SYS_Errorf("ActivateMode: CreateSurface 1 in mode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);

    // upgrade it
    if (initialPrimary->QueryInterface(IID_IDirectDrawSurface3,(LPVOID*)&vidw_lpDDSPrimary) != DD_OK)
        SYS_Errorf("ActivateMode: QueryInterface 1 in mode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);

    // get the back buffer
    ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
    if (initialPrimary->GetAttachedSurface(&ddscaps, &initialBack) != DD_OK)
        SYS_Errorf("ActivateMode: GetAttachedSurface in mode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);

    // upgrade it
    if (initialBack->QueryInterface(IID_IDirectDrawSurface3,(LPVOID*)&vidw_lpDDSBack) != DD_OK)
        SYS_Errorf("ActivateMode: QueryInterface 2 in mode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);

    // release pre-upgraded interfaces
    initialBack->Release();
    initialPrimary->Release();

/*
    // get the offscreen surface
    memset(&ddsdesc, 0, sizeof(ddsdesc));
    ddsdesc.dwSize = sizeof(ddsdesc);
    ddsdesc.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
    ddsdesc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
    ddsdesc.dwWidth = mode->width;
    ddsdesc.dwHeight = mode->height;

    if (vidw_lpDD2->CreateSurface(&ddsdesc, &vidw_lpDDSOffscreen, NULL) != DD_OK)
        SYS_Errorf("ActivateMode: CreateSurface 2 in mode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);

    // lock the offscreen surface and get our current pitch and buffer
    memset(&ddsdesc, 0, sizeof(ddsdesc));
    ddsdesc.dwSize = sizeof(ddsdesc);
    if (vidw_lpDDSOffscreen->Lock(NULL, &ddsdesc, DDLOCK_WAIT, NULL) != DD_OK)
        SYS_Errorf("ActivateMode: Surface Lock in mode %dx%dx%d on device %s(%s)",
            mode->width, mode->height, mode->bitDepth,
            mode->device->name, mode->device->description);	
	vidw_ScreenPitch = ddsdesc.lPitch;
    vidw_ScreenPtr = (byte*)ddsdesc.lpSurface;
    
	// clear the offscreen surface
	for (i=0;i<mode->height;i++)
        memset(vidw_ScreenPtr + i*vidw_ScreenPitch, 0, vidw_ScreenPitch);
*/
    
	// get our other current values
	vidw_ScreenHeight = mode->height;
    vidw_ScreenWidth = mode->width;
    vidw_ScreenBitDepth = mode->bitDepth;

    // plug in the palette if we're in an 8-bit mode
    if (mode->bitDepth == 8)
    {
        for (i=0,palptr=pal,bptr=vidw_Palette; i<256; i++,palptr++,bptr+=3)
        {
            palptr->peRed = bptr[0];
            palptr->peGreen = bptr[1];
            palptr->peBlue = bptr[2];
            palptr->peFlags = 0;
        }
        if (vidw_lpDD2->CreatePalette(DDPCAPS_8BIT | DDPCAPS_ALLOW256,
            pal, &vidw_lpDDPalette, NULL) != DD_OK)
            SYS_Errorf("ActivateMode: CreatePalette in mode %dx%dx%d on device %s(%s)",
                mode->width, mode->height, mode->bitDepth,
                mode->device->name, mode->device->description);
        
        if (vidw_lpDDSPrimary->SetPalette(vidw_lpDDPalette) != DD_OK)
            SYS_Errorf("ActivateMode: SetPalette in mode %dx%dx%d on device %s(%s)",
                mode->width, mode->height, mode->bitDepth,
                mode->device->name, mode->device->description);
    }
}

// functions to map between images and image handles
static vidw_image_t* ImageForId(dword id)
{
    if ((!id) || (id > VIDW_MAXIMAGEHANDLES))
        return(NULL);
    if (!vidw_imagePool[id-1].ddMemSurface)
        return(NULL);
    return(&vidw_imagePool[id-1]);
}

static dword IdForImage(vidw_image_t* img)
{
    if (!img)
        return(0);
    return((dword)(img - vidw_imagePool)+1);
}

static bbool CreateImageSurface(LPDIRECTDRAWSURFACE3* outSurf, dword width, dword height, dword bitDepth, bbool useVideoMemory)
{
    DDSURFACEDESC ddsdesc;
    LPDIRECTDRAWSURFACE initialSurface;

	memset(&ddsdesc, 0, sizeof(ddsdesc));
    ddsdesc.dwSize = sizeof(ddsdesc);
    ddsdesc.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
    if (useVideoMemory)
        ddsdesc.ddsCaps.dwCaps = DDSCAPS_VIDEOMEMORY;
    else
        ddsdesc.ddsCaps.dwCaps = DDSCAPS_SYSTEMMEMORY;
    ddsdesc.dwWidth = width;
    ddsdesc.dwHeight = height;

    if (vidw_lpDD2->CreateSurface(&ddsdesc, &initialSurface, NULL) != DD_OK)
        return(0);
    if (initialSurface->QueryInterface(IID_IDirectDrawSurface3, (LPVOID*)outSurf) != DD_OK)
    {
        initialSurface->Release();
        return(0);
    }
    initialSurface->Release();
    return(1);
}

static bbool RestoreImageSurface(vidw_image_t* img)
{
    DDSURFACEDESC ddsdesc;
    dword i, sPitch, dPitch;
    byte *sPtr, *dPtr;
    
    SYS_Debugf("Restore Required at %f\n", sys_curTime);
    if (img->ddVidSurface == img->ddMemSurface)
        return(0); // system memory surface never needs restoring
    if (img->ddVidSurface->Restore() != DD_OK)
        return(0);
    if (img->flags & VIF_MASKING)
    {
        DDCOLORKEY colorkey;
        colorkey.dwColorSpaceLowValue = colorkey.dwColorSpaceHighValue = 0;
        img->ddVidSurface->SetColorKey(DDCKEY_SRCBLT, &colorkey);
    }
	
    // lock both buffers
    memset(&ddsdesc, 0, sizeof(ddsdesc));
    ddsdesc.dwSize = sizeof(ddsdesc);
    img->ddVidSurface->Lock(NULL, &ddsdesc, DDLOCK_WAIT, NULL);
    dPitch = ddsdesc.lPitch;
    dPtr = (byte*)ddsdesc.lpSurface;
	memset(&ddsdesc, 0, sizeof(ddsdesc));
    ddsdesc.dwSize = sizeof(ddsdesc);
    img->ddMemSurface->Lock(NULL, &ddsdesc, DDLOCK_WAIT, NULL);
    sPitch = ddsdesc.lPitch;
    sPtr = (byte*)ddsdesc.lpSurface;

    // splorf the system memory surface up
    for (i=0;i<img->height;i++)
        memcpy(dPtr+i*dPitch, sPtr+i*sPitch, img->width);

    // unlock the surfaces
    img->ddMemSurface->Unlock(sPtr);
    img->ddVidSurface->Unlock(dPtr);

    return(1);
}

// ddraw enumdevice callback, generates our device list
static BOOL WINAPI EnumDevCallback(GUID FAR* guidPtr,
                                   char* description,
                                   char* name,
                                   void* user)
{        
    if (vidw_numDevices >= VIDW_MAXDEVICES)
        return(DDENUMRET_CANCEL);

    // a NULL guidPtr may be passed in for the primary device
	if (guidPtr)
    {
        vidw_Devices[vidw_numDevices].deviceGuid = *guidPtr;
        vidw_Devices[vidw_numDevices].isNullGuid = 0;
    }
    else
    {
        vidw_Devices[vidw_numDevices].isNullGuid = 1;
    }
    strcpy(vidw_Devices[vidw_numDevices].name, name);
    strcpy(vidw_Devices[vidw_numDevices].description, description);

    // activate the device to get capabilities
	ActivateDevice(&vidw_Devices[vidw_numDevices]);

    // pull in the capabilities structure
	vidw_Devices[vidw_numDevices].devCaps.dwSize = sizeof(DDCAPS);
    vidw_Devices[vidw_numDevices].helCaps.dwSize = sizeof(DDCAPS);
    if (vidw_lpDD2->GetCaps(&vidw_Devices[vidw_numDevices].devCaps,
        &vidw_Devices[vidw_numDevices].helCaps) != DD_OK)
    {
        SYS_Errorf("EnumDevCallback: GetCaps on device %s(%s)",
            vidw_Devices[vidw_numDevices].name,
            vidw_Devices[vidw_numDevices].description);
    }

    // shut down the device
	ActivateDevice(NULL);

    vidw_numDevices++;
    return(DDENUMRET_OK);
}

// ddraw enummode callback, generates our mode list
static BOOL WINAPI EnumModeCallback(DDSURFACEDESC* surfaceDesc,
                                    void* user)
{
    if (vidw_numModes >= VIDW_MAXMODES)
        return(DDENUMRET_CANCEL);
    
    vidw_Modes[vidw_numModes].device = (vidw_device_t*)user;
    vidw_Modes[vidw_numModes].width = surfaceDesc->dwWidth;
    vidw_Modes[vidw_numModes].height = surfaceDesc->dwHeight;
    vidw_Modes[vidw_numModes].bitDepth = surfaceDesc->ddpfPixelFormat.dwRGBBitCount;
    vidw_Modes[vidw_numModes].refresh = surfaceDesc->dwRefreshRate;
    vidw_Modes[vidw_numModes].surfDesc = *surfaceDesc;

    vidw_numModes++;
    return(DDENUMRET_OK);
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================
void VIDW_Init()
{
    dword i;
    
    // enumerate devices
    if (DirectDrawEnumerate((LPDDENUMCALLBACK)EnumDevCallback, NULL) != DD_OK)
        SYS_Errorf("VID_Init failure: DirectDrawEnumerate");
    for (i=0;i<vidw_numDevices;i++)
    {
        // activate the device
		ActivateDevice(&vidw_Devices[i]);

        // set the cooperative level temporarily
		if (vidw_lpDD2->SetCooperativeLevel(sysw_hwnd,
            DDSCL_FULLSCREEN|DDSCL_EXCLUSIVE|DDSCL_NOWINDOWCHANGES) != DD_OK)
            SYS_Errorf("VID_Init failure: SetCooperativeLevel on device %s(%s)",
                vidw_Devices[i].name, vidw_Devices[i].description);

        // enumerate modes in this device 
		if (vidw_lpDD2->EnumDisplayModes(0, NULL, &vidw_Devices[i],
            (LPDDENUMMODESCALLBACK)EnumModeCallback) != DD_OK)
            SYS_Errorf("VID_Init failure: EnumDisplayModes on device %s(%s",
                vidw_Devices[i].name, vidw_Devices[i].description);

        // change the cooperative level back
		if (vidw_lpDD2->SetCooperativeLevel(sysw_hwnd,
            DDSCL_NORMAL) != DD_OK)
            SYS_Errorf("VID_Init failure: SetCooperativeLevel on device %s(%s)",
                vidw_Devices[i].name, vidw_Devices[i].description);

        // shut down the device
		ActivateDevice(NULL);
    }
}

void VIDW_Shutdown()
{
	dword i;

    // free all images
	for (i=1;i<=VIDW_MAXIMAGEHANDLES;i++)
		VIDW_FreeImage(i);

    // setting a null mode will free everything
	ActivateMode(NULL);
}

void VIDW_Frame()
{
    // flip the back buffer and front buffer
    if (vidw_lpDDSPrimary->Flip(NULL, DDFLIP_WAIT)==DDERR_SURFACELOST)
    {
        vidw_lpDDSPrimary->Restore();
        vidw_lpDDSBack->Restore();
        vidw_lpDDSPrimary->Flip(NULL, DDFLIP_WAIT);
    }
}

// image control

// load an image from memory to an image handle
dword VIDW_LoadImage(byte* imageData, sdword width, sdword height, dword bitDepth, bbool masking)
{
    dword i;
    vidw_image_t* img;
    bbool invertV=0;

    // find an available image handle
    for (i=0;i<VIDW_MAXIMAGEHANDLES;i++)
    {
        img = &vidw_imagePool[i];
        if (!img->ddMemSurface)
            break;
    }
    if (i==VIDW_MAXIMAGEHANDLES)
        return(0); // no more images
    if (width < 0)
        width = -width;
    if (height < 0)
    {
        height = -height;
        invertV = 1;
    }
    img->width = width;
    img->height = height;
    img->bitDepth = bitDepth;
    img->flags = 0;
    // FIXME: forces all image surfaces into system memory for the time being, until a video memory prioritizer is added
    if (!CreateImageSurface(&img->ddMemSurface, width, height, bitDepth, 0))
        return(0);
    img->ddVidSurface = img->ddMemSurface;
    if (masking)
    {
        DDCOLORKEY colorkey;
        colorkey.dwColorSpaceLowValue = colorkey.dwColorSpaceHighValue = 0;
        img->ddVidSurface->SetColorKey(DDCKEY_SRCBLT, &colorkey);
        img->flags |= VIF_MASKING;
    }
    if (imageData)
    {
        byte* ptr;
        dword pitch;

        if (ptr = VIDW_LockImage(IdForImage(img), &pitch))
        {
            if (!invertV)
            {
                for (i=0;i<img->height;i++)
                    memcpy(ptr+i*pitch, imageData+i*img->width, img->width);
            }
            else
            {
                for (i=0;i<img->height;i++)
                    memcpy(ptr+i*pitch, imageData+(img->height-i-1)*img->width, img->width);
            }
            VIDW_UnlockImage(IdForImage(img), ptr);
        }
    }
    return(IdForImage(img));
}

// free up a loaded image
void VIDW_FreeImage(dword imageHandle)
{
    vidw_image_t* img;
    if (!(img = ImageForId(imageHandle)))
        return;
    if (img->ddVidSurface != img->ddMemSurface)
        img->ddVidSurface->Release();
    img->ddMemSurface->Release();
    img->ddVidSurface = img->ddMemSurface = NULL;
    img->width = img->height = img->bitDepth = img->flags = 0;
}

// blit an image at the given position
void VIDW_DrawImage(dword imageHandle, sdword scrX, sdword scrY)
{
    vidw_image_t* img;
    RECT r;

    if (!(img = ImageForId(imageHandle)))
        return;

    r.left = r.top = 0;
    r.right = img->width;
    r.bottom = img->height;

    // make sure the blit will be at least partially onscreen
    if ((scrX >= (sdword)vidw_ScreenWidth) || (scrY >= (sdword)vidw_ScreenHeight)
     || (scrX+r.right <= 0) || (scrY+r.bottom <= 0))
        return;

    // clip the dimensions
    if (scrX < 0)
    {
        r.left -= scrX;
        scrX = 0;
    }
    else if (scrX+r.right > (sdword)vidw_ScreenWidth)
    {
        r.right -= (scrX+r.right - (sdword)vidw_ScreenWidth);
    }
    if (scrY < 0)
    {
        r.top -= scrY;
        scrY = 0;
    }
    else if (scrY+r.bottom > (sdword)vidw_ScreenHeight)
    {
        r.bottom -= (scrY+r.bottom - (sdword)vidw_ScreenHeight);
    }

    // blit the thing
    if (vidw_lpDDSBack->BltFast(scrX, scrY, img->ddVidSurface, &r,
        DDBLTFAST_WAIT|((img->flags&VIF_MASKING)?DDBLTFAST_SRCCOLORKEY:DDBLTFAST_NOCOLORKEY)) == DDERR_SURFACELOST)
    {
        RestoreImageSurface(img);
        vidw_lpDDSBack->BltFast(scrX, scrY, img->ddVidSurface, &r,
            DDBLTFAST_WAIT|((img->flags&VIF_MASKING)?DDBLTFAST_SRCCOLORKEY:DDBLTFAST_NOCOLORKEY));
    }
}

// lock access to an image buffer
byte* VIDW_LockImage(dword imageHandle, dword* outImgPitch)
{
    vidw_image_t* img;
    DDSURFACEDESC ddsdesc;
    HRESULT hr;

    if (!(img = ImageForId(imageHandle)))
        return(0);
	
	memset(&ddsdesc, 0, sizeof(ddsdesc));
    ddsdesc.dwSize = sizeof(ddsdesc);
    while ((hr = img->ddVidSurface->Lock(NULL, &ddsdesc, DDLOCK_WAIT, NULL))==DDERR_SURFACELOST)
    {
        vidw_lpDDSPrimary->Restore();
        vidw_lpDDSBack->Restore();
        RestoreImageSurface(img);
    }
    if (hr != DD_OK)
        return(NULL);
    if (outImgPitch)
        *outImgPitch = ddsdesc.lPitch;
    return((byte*)ddsdesc.lpSurface);
}

// release a previous lock
void VIDW_UnlockImage(dword imageHandle, byte* lockPtr)
{
    vidw_image_t* img;
    HRESULT hr;
    
    if (!(img = ImageForId(imageHandle)))
        return;
	while ((hr = img->ddVidSurface->Unlock(lockPtr)) != DD_OK)
    {
        if (hr == DDERR_SURFACELOST)
        {
            vidw_lpDDSPrimary->Restore();
            vidw_lpDDSBack->Restore();
            RestoreImageSurface(img);
        }
    }
}

// image info retrieval functions
dword VIDW_GetImageWidth(dword imageHandle)
{
    vidw_image_t* img;

    if (!(img = ImageForId(imageHandle)))
        return(0);
    return(img->width);
}

dword VIDW_GetImageHeight(dword imageHandle)
{
    vidw_image_t* img;

    if (!(img = ImageForId(imageHandle)))
        return(0);
    return(img->height);
}

dword VIDW_GetImageBitDepth(dword imageHandle)
{
    vidw_image_t* img;

    if (!(img = ImageForId(imageHandle)))
        return(0);
    return(img->bitDepth);
}

bbool VIDW_SetMode(void* mode)
{
    if (!mode)
        return(0);
    ActivateMode((vidw_mode_t*)mode);
    return(1);
}

// change the video mode
bbool VIDW_SetModeEx(dword width, dword height, dword bitDepth)
{
    dword i;
    vidw_mode_t* mode = NULL;

    // see if we have a mode that matches, don't worry about
	// the device for the moment (we'll eventually let users
	// choose the device in-game)
    for (i=0;i<vidw_numModes;i++)
    {
        if ((vidw_Modes[i].width == width)
         && (vidw_Modes[i].height == height)
         && (vidw_Modes[i].bitDepth == bitDepth))
        {
            mode = &vidw_Modes[i];
            break;
        }
    }
	// if there wasn't a matching mode, cut out
    if (!mode)
        return(0);
    
	// go for it
    ActivateMode(mode);

    return(1);
}

// mode info retrieval functions

dword VIDW_GetAvailableModes(void** modeList, dword maxModes)
{
    dword i;

    if (!modeList)
        return(vidw_numModes);
    if (maxModes > vidw_numModes)
        maxModes = vidw_numModes;
    for (i=0;i<maxModes;i++)
        modeList[i] = (void*)&vidw_Modes[i];
    return(vidw_numModes);
}

dword VIDW_GetModeWidth(void* mode)
{
    if (!mode)
        mode = vidw_currentMode;
    return(((vidw_mode_t*)mode)->width);
}

dword VIDW_GetModeHeight(void* mode)
{
    if (!mode)
        mode = vidw_currentMode;
    return(((vidw_mode_t*)mode)->height);
}

dword VIDW_GetModeBitDepth(void* mode)
{
    if (!mode)
        mode = vidw_currentMode;
    return(((vidw_mode_t*)mode)->bitDepth);
}

char* VIDW_GetModeDeviceName(void* mode)
{
    if (!mode)
        mode = vidw_currentMode;
    return(((vidw_mode_t*)mode)->device->name);
}

char* VIDW_GetModeDeviceDescription(void* mode)
{
    if (!mode)
        mode = vidw_currentMode;
    return(((vidw_mode_t*)mode)->device->description);
}

// copy a sequential block of color values into the palette
void VIDW_SetPaletteRange(byte* buffer, dword startColor, dword numColors)
{
    PALETTEENTRY pal[256];
    dword i;
    byte* bptr;
    PALETTEENTRY* palptr;

    // if we're not in an 8-bit video mode, abort
    if ((!vidw_currentMode) || (vidw_currentMode->bitDepth != 8))
        return;

    // if too many colors are being set, abort
    if ((startColor + numColors) > 256)
        return;

    // copy into our internal palette first, for when we change modes
	memcpy(&vidw_Palette[startColor*3], buffer, numColors*3);

    for (i=0,palptr=pal,bptr=buffer; i<numColors; i++,palptr++,bptr+=3)
    {
        palptr->peRed = bptr[0];
        palptr->peGreen = bptr[1];
        palptr->peBlue = bptr[2];
        palptr->peFlags = 0;
    }
    vidw_lpDDPalette->SetEntries(0, startColor, numColors, pal);
}

// set a single palette entry
void VIDW_SetPaletteEntry(byte color, byte r, byte g, byte b)
{
    PALETTEENTRY palentry;

    // if we're not in an 8-bit video mode, abort
    if ((!vidw_currentMode) || (vidw_currentMode->bitDepth != 8))
        return;

    // copy into internal palette first
	vidw_Palette[color*3] = r;
    vidw_Palette[color*3+1] = g;
    vidw_Palette[color*3+2] = b;
    
    vidw_lpDDPalette->GetEntries(0, color, 1, &palentry);
    palentry.peRed = r;
    palentry.peGreen = g;
    palentry.peBlue = b;
    vidw_lpDDPalette->SetEntries(0, color, 1, &palentry);
}

// get the current colors of a palette entry
void VIDW_GetPaletteEntry(byte color, byte *r, byte *g, byte *b)
{
    // if we're not in an 8-bit video mode, abort
    if ((!vidw_currentMode) || (vidw_currentMode->bitDepth != 8))
        return;

    if (r)
        *r = vidw_Palette[color*3];
    if (g)
        *g = vidw_Palette[color*3+1];
    if (b)
        *b = vidw_Palette[color*3+2];
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE VID_WIN.CPP
//**
//****************************************************************************

